// This is the main DLL file.

#include "stdafx.h"

#include "G31DDCAPIWrapper.h"

using namespace G31DDCAPIWrapperSpace;
using namespace System;
using namespace System::Runtime::InteropServices;


//
// ApiHandler
//
HMODULE ourModuleHandle = NULL;
INT32 ourDeviceHandle = -1;
G31DDC_CALLBACKS ourCallbacks;

G31DDC_OPEN_DEVICE ourOpenDevice;
G31DDC_CLOSE_DEVICE ourCloseDevice;
G31DDC_GET_DEVICE_LIST ourGetDeviceList;
G31DDC_IS_DEVICE_CONNECTED ourIsDeviceConnected;
G31DDC_SET_POWER ourSetPower;
G31DDC_SET_DDC1 ourSetDDC1;
G31DDC_SET_FREQUENCY ourSetFrequency;
G31DDC_GET_FREQUENCY ourGetFrequency;
G31DDC_SET_AGC ourSetAGC;
G31DDC_SET_AGC_PARAMS ourSetAGCParams;
G31DDC_SET_DEMODULATOR_MODE ourSetDemodulatorMode;
G31DDC_SET_DEMODULATOR_FILTER_BANDWIDTH ourSetDemodulatorFilterBandwidth;
G31DDC_START_DDC1 ourStartDDC1;
G31DDC_STOP_DDC1 ourStopDDC1;
G31DDC_START_DDC2 ourStartDDC2;
G31DDC_STOP_DDC2 ourStopDDC2;
G31DDC_START_AUDIO ourStartAudio;
G31DDC_STOP_AUDIO ourStopAudio;
G31DDC_SET_CALLBACKS ourSetCallbacks;

G31DDCAPIWrapper::G31DDCAPIWrapper()
{
	instance = this; // Used in callback from managed C++ to C#
	devicesCount = 0;

	IntPtr aPointer = Marshal::StringToHGlobalUni(APILIBRARY);
	ourModuleHandle = LoadLibrary((LPCWSTR)aPointer.ToPointer());

	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("Failed to load {0}", APILIBRARY);
		status = STATUS_APILIBRARY_NOT_FOUND;
		return;
	}

	// Retrieve addresses of used API functions
    ourOpenDevice = (G31DDC_OPEN_DEVICE)GetProcAddress(ourModuleHandle, "OpenDevice");
    ourCloseDevice = (G31DDC_CLOSE_DEVICE)GetProcAddress(ourModuleHandle, "CloseDevice");
	ourGetDeviceList = (G31DDC_GET_DEVICE_LIST)GetProcAddress(ourModuleHandle, "GetDeviceList");
	ourIsDeviceConnected = (G31DDC_IS_DEVICE_CONNECTED)GetProcAddress(ourModuleHandle, "IsDeviceConnected");
	ourSetPower = (G31DDC_SET_POWER)GetProcAddress(ourModuleHandle, "SetPower");
	ourSetDDC1 = (G31DDC_SET_DDC1)GetProcAddress(ourModuleHandle, "SetDDC1");
	ourSetFrequency = (G31DDC_SET_FREQUENCY)GetProcAddress(ourModuleHandle, "SetFrequency");
	ourGetFrequency = (G31DDC_GET_FREQUENCY)GetProcAddress(ourModuleHandle, "GetFrequency");
	ourSetAGC = (G31DDC_SET_AGC)GetProcAddress(ourModuleHandle, "SetAGC");
	ourSetAGCParams = (G31DDC_SET_AGC_PARAMS)GetProcAddress(ourModuleHandle, "SetAGCParams");
    ourSetDemodulatorMode = (G31DDC_SET_DEMODULATOR_MODE)GetProcAddress(ourModuleHandle, "SetDemodulatorMode");
    ourSetDemodulatorFilterBandwidth = (G31DDC_SET_DEMODULATOR_FILTER_BANDWIDTH)GetProcAddress(ourModuleHandle, "SetDemodulatorFilterBandwidth");
	ourSetCallbacks = (G31DDC_SET_CALLBACKS)GetProcAddress(ourModuleHandle, "SetCallbacks");
	ourStartDDC1 = (G31DDC_START_DDC1)GetProcAddress(ourModuleHandle, "StartDDC1");
	ourStopDDC1 = (G31DDC_STOP_DDC1)GetProcAddress(ourModuleHandle, "StopDDC1");
	ourStartDDC2 = (G31DDC_START_DDC2)GetProcAddress(ourModuleHandle, "StartDDC2");
	ourStopDDC2 = (G31DDC_STOP_DDC2)GetProcAddress(ourModuleHandle, "StopDDC2");
	ourStartAudio = (G31DDC_START_AUDIO)GetProcAddress(ourModuleHandle, "StartAudio");
	ourStopAudio = (G31DDC_STOP_AUDIO)GetProcAddress(ourModuleHandle, "StopAudio");

	if (!ourOpenDevice || !ourCloseDevice || !ourGetDeviceList || !ourIsDeviceConnected || !ourSetPower || !ourSetDDC1 || 
		!ourSetFrequency || !ourGetFrequency || 
		!ourSetAGCParams || !ourSetAGC || !ourSetAGCParams || !ourSetDemodulatorMode || !ourSetDemodulatorFilterBandwidth ||
		!ourStartDDC1 || !ourStopDDC1 || !ourStartDDC2 || !ourStopDDC2 || !ourStartAudio || !ourStopAudio ||
		!ourSetCallbacks)
	{
		Console::WriteLine("Incorrect {0}", APILIBRARY);
		FreeLibrary(ourModuleHandle);
		ourModuleHandle = NULL;
		status = STATUS_APILIBRARY_INCORRECT;
		return;
	}

	Console::WriteLine("{0} loaded", APILIBRARY);

	status = STATUS_APILIBRARY_LOADED;
}

G31DDCAPIWrapper::~G31DDCAPIWrapper()
{
	FreeLibrary(ourModuleHandle);
}

DWORD G31DDCAPIWrapper::GetStatus()
{
	return status;
}

BOOL G31DDCAPIWrapper::IsConnected()
{
	if (ourDeviceHandle < 0)
	{
		return FALSE;
	}
	else
	{
		return ourIsDeviceConnected(ourDeviceHandle);
	}
}

String^ G31DDCAPIWrapper::GetDevicePath()
{
	return deviceInfo.devicePath;
}

BYTE G31DDCAPIWrapper::GetInterfaceType()
{
	return deviceInfo.interfaceType;
}

String^ G31DDCAPIWrapper::GetSerialNumber()
{
	return deviceInfo.serialNumber;
}

BOOL G31DDCAPIWrapper::Enumerate()
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);
		return FALSE;
	}

	// Let's work with one device
	G31DDC_DEVICE_INFO aDeviceList[1];
	devicesCount = ourGetDeviceList(aDeviceList, sizeof(G31DDC_DEVICE_INFO));
	if (devicesCount == 0)
	{
		Console::WriteLine("No available device found");
        return FALSE;
	}

	G31DDC_DEVICE_INFO aDeviceInfo = aDeviceList[0];
	deviceInfo.devicePath = gcnew String(aDeviceInfo.DevicePath);
	deviceInfo.interfaceType = aDeviceInfo.InterfaceType;
	deviceInfo.serialNumber = gcnew String(aDeviceInfo.SerialNumber);
	deviceInfo.channelCount = aDeviceInfo.ChannelCount;
	deviceInfo.ddcTypeCount = aDeviceInfo.DDCTypeCount;

	Console::WriteLine("Device #{0} found", deviceInfo.serialNumber);

	return TRUE;
}

BOOL G31DDCAPIWrapper::Open()
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);
		return FALSE;
	}

	if (devicesCount == 0)
	{
		Console::WriteLine("No available device found");
		return FALSE;
	}

	IntPtr aPointer = Marshal::StringToHGlobalAnsi(deviceInfo.serialNumber);
	const char* aSerialNumber = static_cast<char*>(aPointer.ToPointer());
	// Open device
	ourDeviceHandle = ourOpenDevice(aSerialNumber);
	Marshal::FreeHGlobal(aPointer);

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Unable to open device");
		return FALSE;
	}

	Console::WriteLine("Device opened");

	// Set power
    ourSetPower(ourDeviceHandle, TRUE);

	// Example initialization
    ourSetDDC1(ourDeviceHandle, 1); // Set DDC type of DDC1 to 1 (DDC Bandwidth = 24000, SampleRate = 32 kHz, BitsPerSample = 32)
    ourSetAGCParams(ourDeviceHandle, 0, 0.020, 0.500, -15.0);
    ourSetAGC(ourDeviceHandle, 0, TRUE);
    ourSetDemodulatorMode(ourDeviceHandle, 0, G3XDDC_MODE_AM);
    ourSetDemodulatorFilterBandwidth(ourDeviceHandle, 0, 8000);

	return TRUE;
}

BOOL G31DDCAPIWrapper::Close()
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	// Unregister all callback functions
    ourSetCallbacks(ourDeviceHandle, NULL, 0);

    ourSetPower(ourDeviceHandle, FALSE);

	BOOL aResult = ourCloseDevice(ourDeviceHandle);

	if (aResult)
	{
		ourDeviceHandle = -1;

		Console::WriteLine("Device closed");
	}

	return aResult;
}

BOOL G31DDCAPIWrapper::SetFrequency(UINT32 theChannel, UINT32 theFrequency)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourSetFrequency(ourDeviceHandle, theChannel, theFrequency);
}

BOOL G31DDCAPIWrapper::GetFrequency(UINT32 theChannel, UINT32* theFrequency)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourGetFrequency(ourDeviceHandle, theChannel, theFrequency);
}

BOOL G31DDCAPIWrapper::RegisterCallbacks(DDC2PreprocessedStreamCallback^ theDDC2PreprocessedStreamCallback, AudioStreamCallback^ theAudioStreamCallback)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	ddc2PreprocessedStreamCallback = theDDC2PreprocessedStreamCallback; // Callback to C#
	audioStreamCallback = theAudioStreamCallback; // Callback to C#

	// DDC2PreprocessedStream-callback from unmanaged to managed C++
	NotificationDDC2PreprocessedStreamCallbackDelegate^ aDDC2PreprocessedStreamDelegate = gcnew NotificationDDC2PreprocessedStreamCallbackDelegate(NotificationDDC2PreprocessedStreamCallback);
	GCHandle aHandle = GCHandle::Alloc(aDDC2PreprocessedStreamDelegate);
	IntPtr aPointer = Marshal::GetFunctionPointerForDelegate(aDDC2PreprocessedStreamDelegate);
	G31DDC_DDC2_PREPROCESSED_STREAM_CALLBACK aDDC2PreprocessedStreamCallback = static_cast<G31DDC_DDC2_PREPROCESSED_STREAM_CALLBACK>(aPointer.ToPointer());

	// AudioStreamCallback from unmanaged to managed C++
	NotificationAudioStreamCallbackDelegate^ anAudioStreamCallbackDelegate = gcnew NotificationAudioStreamCallbackDelegate(NotificationAudioStreamCallback);
	aHandle = GCHandle::Alloc(anAudioStreamCallbackDelegate);
	aPointer = Marshal::GetFunctionPointerForDelegate(anAudioStreamCallbackDelegate);
	G31DDC_AUDIO_STREAM_CALLBACK anAudioStreamCallback = static_cast<G31DDC_AUDIO_STREAM_CALLBACK>(aPointer.ToPointer());

	// Force garbage collection cycle to prove that the delegate does not get disposed
	GC::Collect();

	// Registration of callbacks
    ourCallbacks.IFCallback = NULL;
    ourCallbacks.DDC1StreamCallback = NULL;
    ourCallbacks.DDC1PlaybackStreamCallback = NULL;
    ourCallbacks.DDC2StreamCallback = NULL;
    ourCallbacks.DDC2StreamCallback = NULL;
    ourCallbacks.DDC2PreprocessedStreamCallback = aDDC2PreprocessedStreamCallback;
    ourCallbacks.AudioStreamCallback = anAudioStreamCallback;
    ourCallbacks.AudioPlaybackStreamCallback = NULL;

	// Register callbacks
    ourSetCallbacks(ourDeviceHandle, &ourCallbacks, 0);

	return TRUE;
}

BOOL G31DDCAPIWrapper::StartDDC1(UINT32 theSamplesPerBuffer)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStartDDC1(ourDeviceHandle, theSamplesPerBuffer);
}

BOOL G31DDCAPIWrapper::StopDDC1()
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStopDDC1(ourDeviceHandle);
}

BOOL G31DDCAPIWrapper::StartDDC2(UINT32 theChannel, UINT32 theSamplesPerBuffer)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStartDDC2(ourDeviceHandle, theChannel, theSamplesPerBuffer);
}

BOOL G31DDCAPIWrapper::StopDDC2(UINT32 theChannel)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStopDDC2(ourDeviceHandle, theChannel);
}

BOOL G31DDCAPIWrapper::StartAudio(UINT32 theChannel, UINT32 theSamplesPerBuffer)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStartAudio(ourDeviceHandle, theChannel, theSamplesPerBuffer);
}

BOOL G31DDCAPIWrapper::StopAudio(UINT32 theChannel)
{
	if (ourModuleHandle == NULL)
	{
		Console::WriteLine("{0} not loaded", APILIBRARY);

		return FALSE;
	}

	if (ourDeviceHandle < 0)
	{
		Console::WriteLine("Device not opened");

		return FALSE;
	}

	return ourStopAudio(ourDeviceHandle, theChannel);
}

void G31DDCAPIWrapper::NotificationDDC2PreprocessedStreamCallback(UINT32 theChannel, const FLOAT* theBuffer, UINT32 theNumberOfSamples, FLOAT theSLevelPeak, FLOAT theSLevelRMS, DWORD_PTR theUserData)
{
	// Callback to C#
	instance->ddc2PreprocessedStreamCallback(theChannel, theBuffer, theNumberOfSamples, theSLevelPeak, theSLevelRMS);
}

void G31DDCAPIWrapper::NotificationAudioStreamCallback(UINT32 theChannel, const FLOAT* theBuffer, const FLOAT* theBufferFiltered, UINT32 theNumberOfSamples, DWORD_PTR theUserData)
{
	// Callback to C#
	instance->audioStreamCallback(theChannel, theBuffer, theBufferFiltered, theNumberOfSamples);
}